// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "Animation.hpp"

namespace Dawn {
  template<typename T>
  struct SimpleKeyframe {
    float_t time;
    T value;
  };

  template<typename T>
  struct SimpleAnimation : public Animation {
    public:
      easefunction_t *easing = &easeLinear;
      T *modifies;
      std::vector<struct SimpleKeyframe<T>> keyframes;

      SimpleAnimation(T *modifies) {
        this->modifies = modifies;
      }

      void addKeyframe(float_t time, T value) {
        struct SimpleKeyframe<T> keyframe;
        keyframe.time = time;
        keyframe.value = value;
        this->duration = mathMax<float_t>(this->duration, time);
        this->finished = false;
        this->keyframes.push_back(keyframe);
      }

      void addSequentialKeyframes(
        float_t startTime,
        float_t frameTime,
        T start,
        T end,
        T step
      ) {
        T v = start;
        float_t n = startTime;
        while(v != end) {
          this->addKeyframe(n, v);
          n += frameTime;
          v += step;
        }
      }

      void addSequentialKeyframes(float_t frameTime, T start, T end) {
        this->addSequentialKeyframes(0, frameTime, start, end, 1);
      }

      void tick(float_t delta) override {
        if(this->finished) return;

        float_t newTime = this->time + delta;

        struct SimpleKeyframe<T> *keyframeNext = nullptr;
        struct SimpleKeyframe<T> *keyframeCurrent = nullptr;
        
        // Find current and next keyframe(s)
        auto itKey = this->keyframes.begin();
        while(itKey != this->keyframes.end()) {
          if(itKey->time > newTime) {
            keyframeNext = &(*itKey);
            break;
          }
          keyframeCurrent = &(*itKey);
          ++itKey;
        }

        // Update values
        if(keyframeCurrent != nullptr && keyframeNext == nullptr) {
          // "End of animation"
          *this->modifies = keyframeCurrent->value;
        } else if(keyframeNext != nullptr) {
          T oldValue;
          float_t oldTime;

          if(keyframeCurrent == nullptr) {
            // "Start of animation"
            oldValue = keyframeCurrent->value;
            oldTime = keyframeCurrent->time;
          } else {
            // "Mid animation"
            oldTime = this->time;
            oldValue = *this->modifies;
          }

          // Slerp between keyframes
          float_t keyframeDelta = this->easing(
            (newTime - oldTime) / (keyframeNext->time - oldTime)
          );
          *this->modifies = oldValue + (
            (keyframeNext->value - oldValue) * keyframeDelta
          );
        }

        // Update time.
        this->time = newTime;

        // Has the animation finished?
        if(newTime < this->duration) return;
        
        // Do we need to loop?
        if(this->loop) {
          this->time = 0;
          return;
        }

        // Animation end.
        this->finished = true;
        this->eventAnimationEnd.invoke();
      }

      void clear() override {
        Animation::clear();
        this->keyframes.clear();
      }
  };
}